@@ -532,15 +532,15 @@ static Bool I830PreInit(ScrnInfoPtr scrn, int flags)
 		return FALSE;
 
 	switch (scrn->depth) {
-	case 8:
 	case 15:
 	case 16:
 	case 24:
 	case 30:
 		break;
+	case 8:
 	default:
 		xf86DrvMsg(scrn->scrnIndex, X_ERROR,
-			   "Given depth (%d) is not supported by I830 driver\n",
+			   "Given depth (%d) is not supported by intel driver\n",
 			   scrn->depth);
 		return FALSE;
 	}
@@ -1093,7 +1093,7 @@ static void I830FreeScreen(FREE_SCREEN_ARGS_DECL)
 	SCRN_INFO_PTR(arg);
 	intel_screen_private *intel = intel_get_screen_private(scrn);
 
-	if (intel) {
+	if (intel && !((uintptr_t)intel & 1)) {
 		intel_mode_fini(intel);
 		intel_close_drm_master(intel);
 		intel_bufmgr_fini(intel);
